/*
   mp_acid_out.c

   Copyright(C) 2002 Gene Savchuk, Fidelis Security Systems
   Copyright(C) 2004,2005 Jonathan Le Plastrier, Farm9.com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: libmp_acid_out.c,v 1.13 2005/06/22 21:12:29 jleplast Exp $ */

#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <time.h>
#include <stdarg.h>

#include <mysql/mysql.h>
#include <mysql/errmsg.h>

#include "mp_export.h"
#include "decode.h"

/* declarations
 */
/* to remember the handler that our host will pass to us
   during initialization
 */
serv_t  *OS;

/* Our host expects these functions
 */
mp_out_init_t mp_out_init;
mp_out_alert_t mp_out_alert;
mp_out_log_t mp_out_log;
mp_out_fini_t mp_out_fini;

static MYSQL BAZA;

static struct			/* parameters */
{
  char *user;
  char *pass;

  char *iface;
  char *host;

  char *server;
  char *db;

  int s_id;
  u_long event_id;
} P;

/* -------- prototypes ---------- */
static int parse_params(char *);

static int select_int(int *null, char *sql, ...);
static int insert(u_long *id, char *sql, ...);
static u_long get_sigid(sid_t *sid, class_t *class, u_int priority);

static char *str_time(time_t timet);
static char *fasthex(u_char *xdata, int length);
static void *ckbuf(char *buf, int *buflen, int minimum);

/* -------- implementations -------- */

/* this function gets executed by dlopen automatically
   Unfortunately, this nice copyright message won't be
   shown in daemon mode.
*/
void _init()
{
  printf(
	 "---------------------------------------------------\n"
	 "           ACID output module for mudpit.\n"
	 "   (c) 2002 G Savchuk, Fidelis Security Systems.\n"
	 "---------------------------------------------------\n"
	 );
  fflush(stdout);
}

int mp_out_init(char *args, serv_t *os)
{
  int schema_version, sensor_cid, event_cid; 

  assert(OS = os);
  if(!parse_params(args)) return 0;

  mysql_init(&BAZA);
  if(!mysql_real_connect(&BAZA, P.server, P.user,
			 P.pass, P.db, 0, NULL,
			 CLIENT_COMPRESS | CLIENT_SSL)) {
    OS->logprint("%s\n", mysql_error(&BAZA));
    return 0;
  }

  if((schema_version = select_int(NULL, "SELECT vseq FROM schema")) != 106)
  {
  OS->fatal("Schema version is %i. Mudpit only supports schema version 106.\n", schema_version);
  return 0;
  }

  if(!P.s_id) {
    u_int s_id = 0;
    if(!P.iface)
      OS->fatal("interface name is needed.\n");

    if(!(s_id = select_int(NULL,
			  "SELECT sid FROM sensor "
			  "WHERE hostname='%s' AND interface='%s' "
			  "AND filter='' AND detail='1' AND encoding='0'",
			  P.host, P.iface))) {
      insert((u_long *)&s_id,
	     "INSERT INTO sensor(hostname, interface, filter, detail, encoding, last_cid) "
	     "VALUES('%s', '%s', '', '1', '0', '0')",
	     P.host, P.iface);
    }

    P.s_id = s_id;
  }

       /* From Snort source: Set the cid value
         * - get the cid value in sensor.last_cid
         * - get the MAX(cid) from event
         * - if snort crashed without storing the latest cid, then
         *     the MAX(event.cid) > sensor.last_cid.  Update last_cid in this case
         */


  sensor_cid = select_int(NULL,
			 "SELECT last_cid FROM sensor WHERE sid='%u'",
			 P.s_id);

  event_cid = select_int(NULL,
			  "SELECT max(cid) FROM event WHERE sid='%u'",
			  P.s_id);

  if ( event_cid > sensor_cid )
     {
	insert(NULL,
		"UPDATE sensor SET last_cid = '%u' WHERE sid = '%u'", 
		event_cid, P.s_id);      
         OS->logprint("database: inconsistent cid information for sid=%u\n",
                        P.s_id);
         OS->logprint("Recovering by rolling forward the cid=%u\n",
                       event_cid);
     }

  P.event_id = event_cid;
  ++(P.event_id); 

#define DL 1
  OS->debug(DL, "mp_acid_out configuration details\n");
  OS->debug(DL, "Detail Level: Always Full\n");
  OS->debug(DL, "Server: %s\n",       P.server);
  OS->debug(DL, "Database:%s\n",      P.db);
  OS->debug(DL, "User: %s\n",         P.user);
  OS->debug(DL, "Password: %s\n",     P.pass);
  OS->debug(DL, "Interface: %s\n",    P.iface);
  OS->debug(DL, "Hostname: %s\n",     P.host);
  OS->debug(DL, "SensorID: %i\n",     P.s_id);
  OS->debug(DL, "EventID:  %i\n",     P.event_id);
#undef DL

  return 1;
}

int mp_out_alert(void *ptr)
{
  UnifiedAlertRecord *uar = (UnifiedAlertRecord *)ptr;
  sid_t *sid = OS->get_sid(uar->event.sig_generator, uar->event.sig_id);
  class_t *class_type = OS->get_class(uar->event.classification);
  u_long acid_sig_id = get_sigid(sid, class_type, uar->event.priority);

  if(!acid_sig_id) OS->fatal("could not get signature id\n");

  insert(NULL,
	 "INSERT INTO event(sid, cid, signature, timestamp) "
	 "VALUES('%u', '%u', '%u', FROM_UNIXTIME('%u'))",
	 P.s_id, P.event_id, acid_sig_id, uar->ts.tv_sec);

  insert(NULL,
	 "INSERT INTO iphdr(sid, cid, ip_src, ip_dst, ip_proto) "
	 "VALUES('%u', '%u', '%u', '%u', '%u')",
	 P.s_id, P.event_id, uar->sip, uar->dip, uar->protocol);

  switch(uar->protocol) {
  case 6:			/* IPPROTO_TCP */
    insert(NULL,
	   "INSERT INTO tcphdr (sid, cid, tcp_sport, tcp_dport) "
	   "VALUES('%u', '%u', '%u', '%u')", 
	   P.s_id, P.event_id, uar->sp, uar->dp);
    break;
  case 17:			/* IPPROTO_UDP */
    insert(NULL,
	   "INSERT INTO udphdr (sid, cid, udp_sport, udp_dport) "
	   "VALUES('%u', '%u', '%u', '%u')", 
	   P.s_id, P.event_id, uar->sp, uar->dp);
    break;
  case 1:			/* IPPROTO_ICMP */
    insert(NULL,
	   "INSERT INTO icmphdr (sid, cid, icmp_type, icmp_code) "
	   "VALUES('%u', '%u', '%u', '%u')", 
	   P.s_id, P.event_id, uar->sp, uar->dp);
    break;
  }

  ++P.event_id;

  return 1;
}

int mp_out_log(void *ptr)
{
  packet_info_t pi;
  UnifiedLogRecord *ulr = (UnifiedLogRecord *)ptr;
  sid_t *sid = OS->get_sid(ulr->log.event.sig_generator, ulr->log.event.sig_id);
  class_t *class = OS->get_class(ulr->log.event.classification);
  u_long acid_sig_id = get_sigid(sid, class, ulr->log.event.priority);

  if(!acid_sig_id) OS->fatal("could not get signature id\n");
  
  insert(NULL,
	 "INSERT INTO event(sid, cid, signature, timestamp) "
	 "VALUES('%u', '%u', '%u', FROM_UNIXTIME('%u'))",
	 P.s_id, P.event_id, acid_sig_id, ulr->log.pkth.ts.tv_sec);

  /* decode the packet */
  if(!decode_packet(&pi, (struct pcap_pkthdr *)&ulr->log.pkth, ulr->pkt))
    OS->logprint("errors decoding captured packet\n"); 

/*   print_payload(&pi); */

  /* save info  */
  if(pi.is_ip) {
    int i;
    insert(NULL,
	   "INSERT INTO iphdr "
	   "(sid, cid, ip_src, ip_dst, ip_ver, "
	   "ip_hlen, ip_tos, ip_len, ip_id, ip_flags, ip_off, "
	   "ip_ttl, ip_proto, ip_csum) "
	   "VALUES ('%u','%u','%lu','%lu','%u',"
	   "'%u','%u','%u','%u','%u','%u',"
	   "'%u','%u','%u')",
	   P.s_id, P.event_id, pi.ip.sip, pi.ip.dip, pi.ip.ver,
	   pi.ip.hl, pi.ip.tos, pi.ip.len, pi.ip.id, pi.ip.flags, pi.ip.off,
	   pi.ip.ttl, pi.ip.proto, pi.ip.csum);

    for(i = 0; i < pi.ip.opt_count; i++) { /* IP options */
      int len = pi.ip.opts[i].len;
      char *t = fasthex(pi.ip.opts[i].data, len > 0 ? len - 2 : 0);
      insert(NULL,
	     "INSERT INTO opt (sid, cid, optid, opt_proto, opt_code, opt_len, opt_data) "
	     "VALUES ('%u','%u','%u','%u','%u','%u','%s') ",
	     P.s_id, P.event_id, i, 0, pi.ip.opts[i].code, len, t);
      free(t);
    }
  }

 /* Log layer 4, non-fragmented only */
  if(pi.is_ip && !pi.ip.off && !pi.ip.mf) {
    if(pi.is_tcp) {
      int i;
      insert(NULL,
	     "INSERT INTO tcphdr "
	     "(sid, cid, tcp_sport, tcp_dport, tcp_seq,"
	     " tcp_ack, tcp_off, tcp_res, tcp_flags, tcp_win,"
	     " tcp_csum, tcp_urp) "
	     "VALUES ('%u','%u','%u','%u','%lu','%lu','%u',"
	     "'%u','%u','%u','%u','%u')",
	     P.s_id, P.event_id, pi.tcp.sp, pi.tcp.dp, pi.tcp.seq,
	     pi.tcp.ack, pi.tcp.off, pi.tcp.x2, pi.tcp.flags, pi.tcp.win,
	     pi.tcp.csum, pi.tcp.urg);

      for(i = 0; i < pi.tcp.opt_count; i++) { /* TCP options */
	int len = pi.tcp.opts[i].len;
	char *t = fasthex(pi.tcp.opts[i].data, len > 0 ? len - 2 : 0);
	insert(NULL,
	       "INSERT INTO opt (sid, cid, optid, opt_proto, opt_code, opt_len, opt_data) "
	       "VALUES ('%u','%u','%u','%u','%u','%u','%s') ",
	       P.s_id, P.event_id, i, 6, pi.tcp.opts[i].code, len, t);
	free(t);
      }
    }

    if(pi.is_udp) {
      insert(NULL,
	       "INSERT INTO udphdr "
	       "(sid, cid, udp_sport, udp_dport, udp_len, udp_csum) "
	       "VALUES ('%u', '%u', '%u', '%u', '%u', '%u')",
	       P.s_id, P.event_id, pi.udp.sp, pi.udp.dp, pi.udp.len,
	       pi.udp.csum);
    }

    if(pi.is_icmp) {
      insert(NULL, 
	     "INSERT INTO icmphdr (sid, cid, icmp_type, icmp_code, icmp_csum) "
	     "VALUES ('%u','%u','%u','%u','%u')",
	     P.s_id, P.event_id, pi.icmp.type, pi.icmp.code, pi.icmp.csum);
      if(pi.icmp.ext) {
	insert(NULL,
	       "UPDATE icmphdr SET icmp_id='%u', icmp_seq='%u' "
	       "WHERE sid='%u' AND cid='%u'",
	       pi.icmp.id, pi.icmp.seq, P.s_id, P.event_id); 
      }
    }
  }

  if(pi.payload && pi.plen > 0) {
    char *hex_payload = fasthex(pi.payload, pi.plen);
    insert(NULL, 
	   "INSERT INTO data(sid, cid, data_payload) "
	   "VALUES('%u', '%u', '%s')",
	   P.s_id, P.event_id, hex_payload);
    free(hex_payload);
  }

  ++P.event_id;

  return 1;
}

void mp_out_fini()
{
   insert(NULL,
            "UPDATE sensor SET last_cid = '%u' WHERE sid = '%u'", 
             P.event_id-1, P.s_id);
  mysql_close(&BAZA);
}

/* ======================================================================== */
static int parse_params(char *args)
{
  if(!args || !*args) {
    OS->logprint("no parameters provided\n"); 
    return 0;
  }

  {
    struct {
      char *name;
      char **str;
      int *num;
#     define STR 1
#     define NUM 2
      int what;
    } conf_pars[] = {
      {"database",  &P.db,     NULL,    STR},
      {"server",    &P.server, NULL,    STR},
      {"user",      &P.user,   NULL,    STR},
      {"password",  &P.pass,   NULL,    STR},
      {"interface", &P.iface,  NULL,    STR},
      {"hostname",  &P.host,   NULL,    STR},
      {"sensor_id", NULL,      &P.s_id, NUM},
      {NULL, NULL, NULL, 0},
    };

    char *tok;

    for(tok = strtok(args, ","); tok; tok = strtok(NULL, ",")) {
      char *arg;
      int i = 0;
      int found = 0;

      while(isspace((int)*tok)) ++tok;
      arg = strsep(&tok, " ");

      for(i = 0; conf_pars[i].name && !found; i++)
	if((strcasecmp(conf_pars[i].name, arg) == 0) && tok) {
	  switch(conf_pars[i].what) {
	  case STR:
	    if(*(conf_pars[i].str)) free(*(conf_pars[i].str));
	    *(conf_pars[i].str) = strdup(tok);
	    break;

	  case NUM:
	    *(conf_pars[i].num) = atoi(tok);
	    break;
	  }

	  found = 1;
	}

	if(!found) OS->logprint("unrecognized parameter \"%s\"\n", arg);
    }
  }

  if(!P.host) {
    char hostname[1024];
    gethostname(hostname, sizeof(hostname));
    P.host = strdup(hostname);
  }

  return 1;
}

/* looks up the acid sig_id for a message and returns it.  If no sig_id exists,
 * one is created
 * XXX:  Unfortunately, the db does not use the same sig_ids that snort does
 */
static u_long get_sigid(sid_t *sid, class_t *class, u_int priority)
{
  u_long sig_id;
  int null = 0;

  assert(sid);

  sig_id = select_int(&null,
		      "SELECT sig_id FROM signature "
		      "WHERE sig_name='%s' AND sig_rev=%u "
		      "AND sig_sid=%u",
		      sid->msg, sid->rev, sid->sid);

  if(null) {			/* Not found - create a new entry */
    int class_id = 0;
    int ref_seq = 0;
    ref_t *ref;
    if(class) {			/* getting/creating class_id */
      class_id = select_int(&null,
			    "SELECT sig_class_id "
			    "FROM sig_class "
			    "WHERE sig_class_name='%s'",
			    class->type);
      if(null) {		/* class_id not found */
	insert((u_long *)&class_id,
	       "INSERT INTO sig_class(sig_class_name) VALUES('%s')", 
	       class->type);
      }
    } /* if(class) */

    insert(&sig_id,		/* creating new signature */
	   "INSERT INTO signature(sig_name, sig_class_id, sig_priority, "
	   "sig_rev, sig_sid) VALUES('%s', '%u', '%u', '%u', '%u')",
	   sid->msg, class_id, priority, sid->rev, sid->sid);

    /* inserting references for the newly-created signature */
    for(ref = sid->ref; ref; ref = ref->next) {
      int ref_system_id = select_int(&null,
				     "SELECT ref_system_id "
				     "FROM reference_system "
				     "WHERE ref_system_name = '%s'",
				     ref->system);
      if(null) {
	insert((u_long *)&ref_system_id,
	       "INSERT INTO reference_system (ref_system_name) "
	       "VALUES ('%s')", ref->system);
      }

      if(ref_system_id > 0) {
	int ref_id = select_int(&null,
				"SELECT ref_id "
				"FROM reference "
				"WHERE ref_system_id = %d AND ref_tag = '%s'",
				ref_system_id, ref->id);
	if(null)
	  insert((u_long *)&ref_id, "INSERT INTO reference (ref_system_id, ref_tag) "
		 "VALUES (%d, '%s')", ref_system_id, ref->id);

	insert(NULL,
	       "INSERT INTO sig_reference (sig_id, ref_seq, ref_id) "
	       "VALUES (%u, %d, %u)", sig_id, ++ref_seq, ref_id);
      }
    } /* for() */
  }

  return sig_id;
}

/* All 'SELECT' queries in this plug-in return integers.
   'null' flag is used to indicate a NULL condition from DB.
 */
int select_int(int *null_flag, char *sql, ...)
{
  va_list ap;
  static int buflen;
  static char *buf;
  MYSQL_RES *res = NULL;
  MYSQL_ROW row;
  int result = 0;

  assert(sql);

  if(!buf) buf = ckbuf(buf, &buflen, 1024);

  while(1) {                    /* allocate as much as neded */
    int len;
    va_start(ap, sql);
    len = vsnprintf(buf, buflen - 1, sql, ap);
    va_end(ap);
    if(len < buflen - 1) break;
    buf = ckbuf(buf, &buflen, len + 1);
  }

  if(null_flag) *null_flag = 0;

  if((mysql_query(&BAZA, buf) == 0) &&
     (res = mysql_store_result(&BAZA))) {

    if(!(row = mysql_fetch_row(res))) {	/* no data */
      if(null_flag) *null_flag = 1;
    }
    else {
      if(!row[0]) {		/* no data */
	if(null_flag) *null_flag = 1;
      }
      else
	result = atoi(row[0]);
    }
  }
  else OS->fatal("%s\n", mysql_error(&BAZA));

  if(res) mysql_free_result(res);

  return result;
}

/* 
 */
static int insert(u_long *id, char *sql, ...)
{
  va_list ap;
  static char *buf;
  static int buflen;

  assert(sql);

  if(!buf) buf = ckbuf(buf, &buflen, 1024);

  while(1) {			/* allocate as much as neded */
    int len;
    va_start(ap, sql);
    len = vsnprintf(buf, buflen - 1, sql, ap);
    va_end(ap);
    if(len < buflen - 1) break;
    buf = ckbuf(buf, &buflen, len + 1);
  }

  if(mysql_query(&BAZA, buf) == 0) {
    if(id) *id = mysql_insert_id(&BAZA);
    {MYSQL_RES *res = mysql_store_result(&BAZA);
    if(res) mysql_free_result(res);}
  }
  else OS->fatal("%s\n", mysql_error(&BAZA));

  return 1;
}

/* 
 */
static char *str_time(time_t timet)
{
  static char buf[256];
  struct tm *tm = localtime(&timet);

/*     tm = gmtime(&timet); */

  strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", tm);

  return buf;
}

static char *fasthex(u_char *xdata, int length)
{
    char conv[] = "0123456789ABCDEF";
    char *retbuf = NULL; 
    char *index;
    char *end;
    char *ridx;

    index = xdata;
    end = xdata + length;
    retbuf = (char *) calloc((length*2)+1, sizeof(char));
    ridx = retbuf;

    while(index < end)
    {
        *ridx++ = conv[((*index & 0xFF)>>4)];
        *ridx++ = conv[((*index & 0xFF)&0x0F)];
        index++;
    }

    return retbuf;
}

/* allocates at least the requested memory chunk
 */
static void *ckbuf(char *buf, int *buflen, int minimum)
{
  assert(buflen);

  if(*buflen < minimum) {
    int newlen = 1024 * (minimum/1024 + 1);
    if(buf) free(buf);
    assert(buf = malloc(*buflen = newlen));
  }

  return buf;
}
