/*
   mp_conf.c

   Copyright(C) 2002 Gene Savchuk, Fidelis Security Systems

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: mp_conf.c,v 1.2 2004/12/11 19:21:03 jleplast Exp $ */

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <paths.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "mp_conf.h"
#include "mp_spoolman.h"
#include "mp_util.h"

/* =================================================== */

static void parse_mudpit_cf(char *filename);
static char *fix(char **filename, char *def_dir, char *def_file);
static void usage();

/* =================================================== */

/* Global functions
 */

/* Parses command line and sets/verifies the main config file name.
 */
void parse_args(int argc, char **argv)
{
  static int second_time = 0;
  int ch;                        /* storage var for getopt info */
  extern char *optarg;
  int longind;

  struct option longopts[] = {
    { "daemon", 0, 0, 'D' },
    { "help", 0, 0, 'h' },
    { "once",   0, &GCF.run_once, 1 },
    { "nice",   1, 0, 'n' },
    {0,0,0,0},
  };

  debug(V_func, "-> parse_args()\n");

  if(!second_time) second_time = 1;
  else
    fatal("Command line parameters are usually parsed only once.\n");

  while((ch = getopt_long(argc, argv, "n:c:vhD", longopts, &longind)) != -1) {
    switch(ch) {
    case 'c':
      GCF.config_file = strdup((const char *)optarg);
      break;
    case 'v': /* increment verbosity */
      GCF.verbose++;
      break;
    case 'n': /* nice level */
      GCF.nice = atoi(optarg);
      break;
    case 'D': /* run in daemon mode */
      GCF.daemon_flag = 1;
      break;
    case 0: break;
    default:
      usage();
      exit(1);
    }
  }

  /* checking/setting configuration file */
  {
    char config_dir[MAXPATHLEN];
    if(!GCF.config_file)
      GCF.config_file = strdup("/etc/" PACKAGE_NAME ".conf");

    check_file(GCF.config_file, 0);

    /* getting config dir from config_file variable */
    if(snprintf(config_dir, sizeof(config_dir), "%s", GCF.config_file) >=
       sizeof(config_dir))
      fatal("config filename buffer overflow\n");

    *(strrchr(config_dir, '/') + 1) = 0; /* preserve the trailing '/' */

    check_dir(config_dir, 0);

    GCF.config_dir = strdup(config_dir);
  }
}

/* parses config file and does sanity checking
 */
void parse_conf()
{
  debug(V_all, "-> parse_conf(\"%s\")\n", GCF.config_file);

  parse_mudpit_cf(GCF.config_file);

  /* Setting default values, checking for inconsistencies, etc */
  check_file(fix(&GCF.sid_file,
		 GCF.config_dir, "sid-msg.map"), 0);

  check_file(fix(&GCF.gen_file,
		 GCF.config_dir, "gen-msg.map"), 0);

  check_file(fix(&GCF.class_file,
		 GCF.config_dir, "classification.config"), 0);

  check_file(fix(&GCF.urlhead_file,
		 GCF.config_dir, "reference.config"), 0);

  /* assign pid file only if we don't have one already */
  if(!GCF.pid_file) {
    char buf[MAXPATHLEN];

    if(snprintf(buf, sizeof(buf), _PATH_VARRUN PACKAGE_NAME ".pid") > sizeof(buf) - 1)
      fatal("pid filename buffer overflow\n");

    GCF.pid_file = strdup(buf);
  }

  if(!SPOOLS_NUMBER) fatal("has nothing to do.\n");
  else { /* check/process spool parameters */
    int i;
    for(i = 0; i < SPOOLS_NUMBER; i++) {
      spool_t *s = &SPOOLS[i];
      out_mod_t *om = s->o_mods;

      if(!om)
	fatal("spool \"%s\": no output modules specified\n", s->spool_dir);

      check_dir(s->spool_dir, 1);
      if(s->arch_dir) check_dir(s->arch_dir, 1);

      for(; om; om = om->next)
	check_file(om->name, 0);

      fix(&s->check_file, s->spool_dir, "checkpoint");

      s->restarts = 0;
    }
  }
}

void print_conf(int argc, char **argv) 
{
#define m(x) (x)?x:n
#define sf "\t\"%s\"\n"
#define uf "\t%u\n"
  int i;
  char *n = "<null>";

  printf(">>> Configuration options <<< \n");
  printf("config dir:"     sf, m(GCF.config_dir));
  printf("config file:"    sf, m(GCF.config_file));
  printf("class file:"     sf, m(GCF.class_file));
  printf("sid file:"       sf, m(GCF.sid_file));
  printf("gen file:"       sf, m(GCF.gen_file));
  printf("URL file:"       sf, m(GCF.urlhead_file));
  printf("pid file:"       sf, m(GCF.pid_file));
  printf("verbosity:"      uf, GCF.verbose);
  printf("daemon:"         uf, GCF.daemon_flag);
  printf("run_once:"       uf, GCF.run_once);
  printf("nice:"           uf, GCF.nice);

  printf("Spool:\n");
  for(i = 0; i < SPOOLS_NUMBER; i++) {
    out_mod_t *t;
    printf("spool dir:"     sf, m(SPOOLS[i].spool_dir));
    printf("archive dir:"   sf, m(SPOOLS[i].arch_dir));
    printf("checkpoint:"    sf, m(SPOOLS[i].check_file));
    printf("lock:"          sf, m(SPOOLS[i].lock));
    printf("uid:"           uf,   SPOOLS[i].uid);
    printf("\toutput:\n");
    for(t = SPOOLS[i].o_mods; t; t = t->next) {
      printf("\tname:"        sf, m(t->name));
      printf("\tparams:" sf, m(t->params));
      printf("\n");
    }
    printf("delete_processed:" uf, SPOOLS[i].delete_processed);
    printf("--\n");
  }

  printf("command line: ");
  for(i = 0; i < argc; i++) {
    printf("%s ", argv[i]);
  }

  printf("\n");
  printf(">>> -------------------- <<< \n\n");
  fflush(stdout);

#undef m
#undef sf
#undef uf
}

/*----------------------- Local Functions --------------------*/

static void parse_mudpit_cf(char *filename)
{
  extern int yyparse(void);
  extern FILE *yyin, *yyout;
  extern int yylineno;
  int error_while_parse = 0;

  assert(filename);

  yylineno = 0;
  if((yyin = fopen(filename, "r"))) {
    yyout = fopen("/dev/null", "w");
    error_while_parse = yyparse();

    fclose(yyin);
    if(yyout) fclose(yyout);
  }
  else
    fatal("open %s: %s\n", filename, strerror(errno));

  if(error_while_parse)
    fatal("%s:%d: parsing error\n", filename, yylineno);
}

/* Check and corrects, if needed, the following:
   1) filename is not null
   2) filename starts from '/'
   def_dir and def_file are used, if necessary.
 */
static char *fix(char **filename, char *def_dir, char *def_file)
{
  char buffer[MAXPATHLEN];

  assert(filename);
  assert(def_dir && def_file);

  if(!*filename) {
    if(snprintf(buffer, sizeof(buffer) - 1, "%s/%s",
		def_dir, def_file) >= sizeof(buffer))
      fatal("fixing \"%s\": buffer overflow\n", def_file);
    *filename = strdup(buffer);

    debug(V_gen, "\"%s\": default name is assigned.\n",
	  *filename);
  }

  else if(**filename != '/') {
    if(snprintf(buffer, sizeof(buffer) - 1, "%s/%s",
		def_dir, *filename) >= sizeof(buffer))
      fatal("%s config buffer overflow\n", def_file);

    free(*filename);
    *filename = strdup(buffer);

    debug(V_gen, "\"%s\": default directory is assigned.\n",
	  *filename);
  }

  return *filename;
}

static void usage()
{
  printf("USAGE: %s [options]\n", PACKAGE_NAME);
  printf("options:\n");
  printf("-c <config file>\n");
  printf("-v [-v [-v]...]  increase verbosity.\n");
  printf("-D|--daemon      daemon mode.\n");
  printf("--once           process spools once and exit.\n");
  printf("-n|--nice level  set priority level.\n");
  printf("-h|--help        this info.\n");

  return;
}
