/*
   mp_spoolman.c

   Copyright(C) 2002 Gene Savchuk, Fidelis Security Systems
   Copyright(C) 2004,2005 Jonathan Le Plastrier, Farm9.com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* $Id: mp_spoolman.c,v 1.9 2005/06/22 21:12:29 jleplast Exp $ */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <dirent.h>
#include <sys/stat.h>
#include <setjmp.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>
#include <dlfcn.h>

/* FARM9_BSD_PORT Added to work under FreeBSD */
#include <sys/resource.h>

#include "mp_conf.h"
#include "mp_spoolman.h"
#include "mp_util.h"
#include "unified.h"

typedef void (sighandler_t)(int);

/* allocated from configparse.c and parser.y */
spool_t *SPOOLS = NULL;		/* the global spool array */
int SPOOLS_NUMBER = 0;		/* the total number of spools allocated */

/* Static declarations */
static char *spool_name[2] = {"alert", "log"};
static u_int32_t spool_magic[2] = {ALERT_MAGIC, LOG_MAGIC};

static sighandler_t child_handler; /* PARENT */

static void start_em();
static int catch_em(sigset_t *);

static pid_t start_the_spool(spool_t *); /* PARENT->CHILD */
static void out_send(spool_t *sph, void *uar, void *ulr); /* CHILD */
static int process_the_spool(spool_t *); /* CHILD */

static int check_magic(char *dirname, char *filename, int num);	/* CHILD */

static int get_alert(UnifiedAlertRecord *, spool_t *); /* CHILD */
static int get_log(UnifiedLogRecord *, spool_t *, char *packet, int p_len); /* CHILD */
static int read_piece(char *buffer, int len, spool_t *sph, int num, int open_new); /* CHILD */

static int read_header(spool_t *sph, int num); /* CHILD */

static void delete_and_arch(char *filepath, int delete_processed, char *arch_dir); /* CHILD */
static void close_spool_file(spool_t *sph, int num); /* CHILD */
static int find_spool_file(spool_t *sph, int num); /* CHILD */

static int save_check(spool_t *); /* CHILD */
static void restore_check(spool_t *); /* CHILD */

typedef struct {
  pid_t pid;
  struct timeval start_time;		/* time of start */
} proc_info_t;

static proc_info_t *spool_proc;

/* Public implementations */

/* starts separate processes for each spool directory and watches them.
 */
int process_spools(void)	/* PARENT */
{
  int i;
  sigset_t set, oldset;
  pid_t pid;

  sigfillset(&set);	   /* the intention is to block all signals */
  sigprocmask(SIG_SETMASK, &set, &oldset);

  spool_proc = my_alloc(sizeof(proc_info_t) * SPOOLS_NUMBER);

  /*
    this is where the parent process gets himself into an endless loop.
    there are three possibilities to escape from it:
    1) to get an unrecoverable child error (exit code CHILD_FATAL).
    2) to get a signal (see sighandler() in mudpit.c for more info).
    3) to be in run_once mode.

    The number of restarts for each spool is tracked,
    but not really used for now.
  */
  do {
    start_em();
    if(!catch_em(&oldset)) return 0;

  } while(!GCF.run_once);	/* cycle here if NOT in run_once mode */

  sigprocmask(SIG_SETMASK, &oldset, NULL); /* let'em come */
  return 1;
}

/* kills spool processors and frees spool handlers.
*/
void cleanup_spools(void)		/* PARENT */
{
  int i;
  debug(V_func, "-> cleanup_spools(%d)\n", SPOOLS_NUMBER);

  if(spool_proc) {
    int still_running = 0;	/* refreshing spool_proc array */
    for(i = 0; i < SPOOLS_NUMBER; i++) {
      if(spool_proc[i].pid)
	if(waitpid(spool_proc[i].pid, NULL, WNOHANG) != spool_proc[i].pid)
	  still_running++;
	else spool_proc[i].pid = 0;
    }

    if(still_running) {
      sighandler_t *oldsig = signal(SIGTERM, SIG_IGN); /* anti-suicidal action */
      logprint("sending TERM signal to spool processors.\n");
      kill(0, SIGTERM);		/* kill the process group */
      signal(SIGTERM, oldsig);

      for(i = 0; i < SPOOLS_NUMBER; i++) /* finish 'em one by one nicely */
	if(spool_proc[i].pid) {
	  if(waitpid(spool_proc[i].pid, NULL, WNOHANG) != spool_proc[i].pid) {
	    logprint("sending TERM signal to process [%d]\n", spool_proc[i].pid);
	    kill(spool_proc[i].pid, SIGTERM);
	  }
	}

      for(i = 2; i = sleep(i); ); /* try to sleep for couple of seconds */

      for(i = 0; i < SPOOLS_NUMBER; i++) /* finish 'em one by one with a vengeance */
	if(spool_proc[i].pid) {
	  if(waitpid(spool_proc[i].pid, NULL, WNOHANG) != spool_proc[i].pid) {
	    logprint("sending KILL signal to process [%d]\n", spool_proc[i].pid);
	    kill(spool_proc[i].pid, SIGKILL);
	  }
	}

    } /* if(still_running) */
    free(spool_proc);
    spool_proc = NULL;
  }

  for(i = 0; i < SPOOLS_NUMBER; i++) {
    spool_t *sph = &SPOOLS[i];

    if(sph->spool_dir) {
      logprint("closing spool \"%s\"\n", sph->spool_dir);
      free(sph->spool_dir);
    }
    if(sph->arch_dir)     free(sph->arch_dir);
    if(sph->check_file)   free(sph->check_file);
    if(sph->lock)         free(sph->lock);

    while(sph->o_mods) {
      out_mod_t *next = sph->o_mods->next;

      if(sph->o_mods->name) free(sph->o_mods->name);
      if(sph->o_mods->params) free(sph->o_mods->params);
      free(sph->o_mods);

      sph->o_mods = next;
    }
  }

  if(SPOOLS) free(SPOOLS);
  SPOOLS = NULL;
  SPOOLS_NUMBER = 0;
}

/* Static implementations */

/* starts spools in cycle
 */
void start_em()
{
  int i;
  pid_t pid;

  for(i = 0; i < SPOOLS_NUMBER; i++) { /* start/restart spool processors */
    if(!spool_proc[i].pid) {
      struct timeval tt;
      double tm_diff;

      gettimeofday(&tt, NULL);
      tm_diff = tv_diff(&tt, (struct timeval *)&spool_proc[i].start_time);

      if((tm_diff) < 10) { /* do not restart immediately */
	int secs = 11 - (int)tm_diff;
	logprint("spool \"%s\" is disabled for %d seconds.\n",
	    SPOOLS[i].spool_dir, secs);

	alarm(secs); /* SIGALRM'll remind us about this dead spool processor */
	continue;
      }

      while((pid = start_the_spool(&SPOOLS[i])) == -1) {
	logerr("start_the_spool", NULL);
	my_sleep();		/* <-- waiting for better times here */
      }

      spool_proc[i].pid = pid;
      gettimeofday((struct timeval *)&spool_proc[i].start_time, NULL);

      SPOOLS[i].restarts++;	/* some statistics */
    }
  }
}

/* catch dead spools
 */
int catch_em(sigset_t *oldset)
{
  do { /* in run_once mode it will cycle until all spool processors are finished */
    int still_running = 0;	/* <-- this flag makes sence in run_once mode only */
    int pid;
    int status;
    int i;

    sigsuspend(oldset); /* <-- the only place we allow signals in process_spool() */

    while((pid = waitpid(0, &status, WNOHANG)) > 0) {
      for(i = 0; i < SPOOLS_NUMBER; i++) {
	if(spool_proc[i].pid == pid) {
	  if(WIFEXITED(status)) {
	    int stat = WEXITSTATUS(status);

	    if(stat == CHILD_FATAL) {
	      logprint("spool [%d] has unrecoverable problems\n", pid);
	      return 0;
	    }
	    else
	      logprint("spool [%d] has exited with code %d\n", pid, stat);
	  }
	  else if(WIFSIGNALED(status))
	    logprint("spool [%d] has got signal %d (%s)\n",
		pid, WTERMSIG(status), strsignal(WTERMSIG(status)));

	  spool_proc[i].pid = 0;
	  break;
	}
      }
    }

    if(GCF.run_once) {
      for(i = 0; i < SPOOLS_NUMBER; i++)
	if(spool_proc[i].pid) still_running = 1;

      if(!still_running) break; /* spool processors are finished */
    }

  } while(GCF.run_once);	/* cycle here if in run_once mode */

  return 1;
}

/* Exit flag for process_the_spool */
static int spool_exit = 0; /* XXX is this unique per process? */
static void spool_exit_handler(int sig) { spool_exit = 1; }

/* prepares environment in newly-created process,
   loads output modules and calls process_the_spool()
   to do the rest
 */
static pid_t start_the_spool(spool_t *sph) /* PARENT -> CHILD */
{
  pid_t pid;

  assert(sph);

  if((pid = fork())) return pid; /* parent/error */

  else {			/* child */
    sigset_t set;
    out_mod_t *out;
    serv_t *os;

    close_pidfile();

    {int isig = 0; while(isig < NSIG) signal(isig++, SIG_DFL);}
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);

    /* set signal handler to spool_exit_handler() */
    signal(SIGTERM, spool_exit_handler);
    /* Ignore SIGINTs so that CTRL-C will not kill the children */
    /* Instead we let the parent do that so that fini() gets called */
    signal(SIGINT, SIG_IGN);

    assert(os = init_services());

    if(GCF.nice)
      if(setpriority(PRIO_PROCESS, 0, GCF.nice) == -1)
	logerr("setpriority", NULL);

    if(getuid() == 0 && sph->uid != 0) {/* We are root and gonna change this */
      if(setregid(sph->gid, sph->gid) == -1)
	fatal("setregid(%d): %s\n", sph->gid, strerror(errno));
      if(setreuid(sph->uid, sph->uid) == -1)
	fatal("setreuid(%d): %s\n", sph->uid, strerror(errno));
    }

    for(out = sph->o_mods; out; out = out->next) {
      void *handle = NULL;
      mp_out_init_t *out_init = NULL;
      mp_out_fini_t *out_fini = NULL;

      debug(V_gen, "initializing module \"%s\"\n", out->name);

      if(!(handle = dlopen(out->name, RTLD_NOW)))
	fatal("spool \"%s\": %s\n", sph->spool_dir, dlerror());

      if(!(out_init = dlsym(handle, OUT_INIT))) /* init is mandatory */
	fatal("spool \"%s\": no '%s' function in \"%s\"\n",
	      sph->spool_dir, OUT_INIT, out->name);

      out->out_log = dlsym(handle, OUT_LOG);
      out->out_alert = dlsym(handle, OUT_ALERT);

      if(!out->out_log && !out->out_alert) /* should have at least one */
	fatal("spool \"%s\": neither '%s' nor '%s' exists in \"%s\"\n",
	      sph->spool_dir, OUT_ALERT, OUT_LOG, out->name);

      out->out_fini =  dlsym(handle, OUT_FINI); /* optional */

      while(!out_init(out->params, os)) { /* forse the success */
	logprint("error initializing \"%s\": retrying\n", out->name);
	my_sleep();
      }

      if(out->out_fini) atexit(out->out_fini);
    } /* for(;;) */

    logprint("spool \"%s\" has %sstarted. uid/gid = %d/%d\n",
	sph->spool_dir, sph->restarts > 1 ? "re" : "",
	getuid(), getgid());

    /* Enter normal run state */
    exit(process_the_spool(sph));

  } /* else */

  return 0;			/* unreachable */
}

/* setting our own timeout for out_send():
   There is something in the MySQL that sometimes
   takes forever to accomplish.
 */
static sigjmp_buf CTX;
static void sig_handler(int sig) {siglongjmp(CTX, 1);}

/* sends alert/log to output modules, one transfer per event.
   LOG data has higher priority.
 */
static void out_send(spool_t *sph, void *uar, void *ulr)
{
  out_mod_t *o;
  static int tries = 3;

  assert(sph);

  for(o = sph->o_mods; o; o = o->next) {
    int error = 0, result = 0;

    do {
      if(0 == sigsetjmp(CTX, 1)) {
	alarm(60);		/* 1-minute timeout */
	signal(SIGALRM, sig_handler);

	{int fd = spool_lock(sph->lock);

	if(ulr && o->out_log)	/* try LOG first */
	  result = o->out_log(ulr);

	else if(uar && o->out_alert)
	  result = o->out_alert(uar);
	
	else result = 1;	      /* did nothing - no reason for retry. */

	spool_unlock(fd);}
	alarm(0);
      }
      else child_fatal("timeout writing to module \"%s\"\n", o->name);

      if(!result) {
	if(++error >= tries)
	  child_fatal("exiting after %d tries\n", tries);
	logprint("error writing to module \"%s\", retrying\n", o->name);
	my_sleep();
      }

    } while(!result);
  }

}

/* this is the spool processor. It picks up events from ALERT and LOG
   in synchronized manner and calls out_send()
   If SIGTERM received need to finish and return to caller
 */
int process_the_spool(spool_t *sph)
{
# define A_id uar.event.event_id
# define L_id ulr.log.event.event_id

  UnifiedAlertRecord uar;
  UnifiedLogRecord ulr;
  char packet[8192];

  int diff;
  struct timeval l_tv, a_tv;
  int read_log = 1, read_alert = 1;
  int log_empty = 0;

  assert(sph);

  debug(V_func, "-> process_the_spool()\n");

  /* restore checkpoint. If not successful, new file will be
     opened automatically during the first read_piece() call */
  restore_check(sph);

  while(!spool_exit) {
    if(read_alert) {
      register int aret;
      tv_zero(&a_tv);

      if((aret = get_alert(&uar, sph)) == 0) return 0;

      else if(aret == -1) debug(V_func, "alert: no more data\n");

      else tv_copy(&a_tv, &uar.ts);
    }
    else read_alert = 1;	/* reset read_alert */

    if(read_log) {
      register int lret;
      tv_zero(&l_tv);

      if((lret  = get_log(&ulr, sph, packet, sizeof(packet))) == 0) return 0;

      else if(lret == -1) debug(V_func, "log: no more data\n");

      else tv_copy(&l_tv, &ulr.log.pkth.ts);
    }
    else read_log = 1;		/* reset read_log */

    if(!save_check(sph))		/* checkpoint */
      child_fatal("could not save checkpoint.\n");


/* Check what we've got, case by case
 */
    if(GCF.run_once && tv_is_null(&a_tv) && tv_is_null(&l_tv))
      return 1;

/* Case 1: empty ALERT.
   Most probably there is no more data available:
   wait and try again.
*/
    if(tv_is_null(&a_tv)) {
      debug(V_func, "alert is empty\n");

      if(!tv_is_null(&l_tv)) read_log = 0; /* preserve the LOG */
      else debug(V_func, "and log too\n");

      my_sleep();

      continue;
    }
 
/* Case 2: no LOG.
   Try to read again. If still unsuccessful, send ALERT out.
*/
    if(tv_is_null(&l_tv)) {	/* no LOG available */
      if(log_empty) {		/* still unavailable: write ALERT */
	if(!tv_is_null(&a_tv)) out_send(sph, &uar, NULL);	/* write ALERT <----------------  */
      }
      else {			/* try to read the LOG again */
	log_empty ++;
	read_alert = 0;
	my_sleep();
      }

      continue;
    }
    else log_empty = 0;

/* Case 3: found both. compare times and decide what to.
 */
    /* equal */
    if((diff = tv_compare(&a_tv, &l_tv)) == 0) { /* generated out of the same event */

      if(A_id == L_id)
	out_send(sph, &uar, &ulr); /* write BOTH and let out_send() choose <-----------------  */

      else { /* try to recover: read the stream with lesser Event id */
	logprint("same timestamps but different Event IDs (%d, %d)\n", A_id, L_id);
	if(A_id < L_id) read_log = 0;
	else read_alert = 0;
      }
    }

/* ALERT is earlier: send it out and block LOG reading for the next cycle */
    else if(diff < 0) {
      out_send(sph, &uar, NULL); /* write ALERT <--------------  */
      read_log = 0;
    }

/* LOG is earlier: should not happen.
   Block ALERT reading until alert <= log.
 */
    else if(diff > 0) {
      debug(V_func, "log event is earlier, strange. read the next one\n");
      read_alert = 0;
    }
  }

  return 1;
}

/*
  reads 'magic' from file.
  returns:
  1 if it's the right file
  0 if it's not.
*/
static int check_magic(char *dirname, char *filename, int num)
{
  char fname[MAXPATHLEN];
  int result = 0;

  assert(dirname);
  assert(filename);

  if(snprintf(fname, sizeof(fname), "%s/%s", dirname, filename) >= sizeof(fname))
    fatal("Buffer overflow.\n");
  else {
    int fd = open(fname, O_RDONLY);
    if(fd == -1)
      logerr("open", fname);

    else {
      u_int32_t magic;
      if(read(fd, &magic, sizeof(magic)) == sizeof(magic))
	result = (magic == spool_magic[num]);

      close(fd);
    }
  }

  debug(V_func, "file '%s' is%s a %s file\n",
	  fname, result ? "" : " not", spool_name[num]);

  return result;
}

/*
returns:
0: error
1: success
*/
static int read_header(spool_t *sph, int num)
{
  spoolfile_t *f;
  UnifiedAlertFileHeader alert_header;
  UnifiedLogFileHeader log_header;

  char *bufs[2] = {(char *)&alert_header, (char *)&log_header};
  int lens[2] = { sizeof(alert_header), sizeof(log_header) };

  assert(sph);
  debug(V_func, "-> read_header(%d)\n", num);

  f = &sph->file[num];

  if(read(f->fd, bufs[num], lens[num]) != lens[num]) {
    logerr("read", f->filepath);
    return 0;
  }

  f->offset = lens[num];

  {
    u_int32_t c_m[2] = {ALERT_MAGIC, LOG_MAGIC};
    u_int32_t m[2] =   {alert_header.magic, log_header.magic};
    if(c_m[num] != m[num]) {
      logprint("Corrupted %s file \"%s\"\n", spool_name[num], f->filepath);
      return 0;
    }
  }

  return 1;
}

/*
This function opens a spool file if there is no
one opened already, or if there is one newer than
currently opened.
If there is no file found and no file opened already,
function returns 2 (no change since last check)
Retuns:
 0: error
 1: success
-1: current file has not changed since last check
    (interpreted as EOF)
*/
static int find_spool_file(spool_t *sph, int num)
{
  char fname[MAXPATHLEN];

  spoolfile_t *f;
  time_t min_timet = 0;
  struct stat statbuf;

  assert(sph);
  debug(V_func, "-> find_spool_file()\n");

  f = &sph->file[num];

  /* Step ONE: check for changes in the currently opened file */

  if(f->fd && f->timet) {
    if(fstat(f->fd, &statbuf) == -1)
      child_fatal("stat \"%s\": %s\n", f->filepath, strerror(errno));

    if(f->inode != statbuf.st_ino)
      child_fatal("file \"%s\" has been replaced\n", f->filepath);
    if(f->size > statbuf.st_size)
      child_fatal("file \"%s\" has been truncated\n", f->filepath);

    f->size = statbuf.st_size;

    /* if the file was updated since last time, continue to use it */
    if(f->size > f->offset) {
      debug(V_gen, "using updated %s file '%s', offset %u, size %u\n",
	    spool_name[num], f->filepath, f->offset, f->size);

      return 1;
    }
  }

  /* Step TWO: find a new (or newer) file */

  {
    struct dirent *entry;
    DIR *dir;

    if(!(dir = opendir(sph->spool_dir)))
      child_fatal("open directory \"%s\": %s\n",
		  sph->spool_dir, strerror(errno));

    
    while((entry = readdir(dir))) {
      if(!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, "..")) continue;
      if(check_magic(sph->spool_dir, entry->d_name, num)) {
	/* this is a file we may want - get the extention */
	char *ext = strrchr(entry->d_name, '.');

	if(ext) {
	  time_t we_may_want = strtoul(ext + 1, NULL, 10);

	  if(we_may_want > f->timet) {
	    if(min_timet == 0 || we_may_want < min_timet) {
	      min_timet = we_may_want;
	      strncpy(fname, entry->d_name, sizeof(fname));
	    }
	  }
	}
      }
    }
    closedir(dir);
  }

  if(!min_timet) {	/* No file found. Why? */
    if(!f->timet)	       /* because there are no files at all */
      debug(V_gen, "spool directory \"%s\" is emty\n", sph->spool_dir);

    else {		/* because there is no file newer than we have opened */
      /* check for length's changes again */
      if(fstat(f->fd, &statbuf) == -1)
	child_fatal("stat \"%s\": %s\n", f->filepath, strerror(errno));

      f->size = statbuf.st_size;

      debug(V_gen, "using %s file \"%s\", offset %u, size %u\n",
	    spool_name[num], f->filepath, f->offset, f->size);
 
      if(f->size > f->offset) return 1;
    }
    return -1;			/* EOF */
  }

  /* Step THREE: Open the newfound file */

  debug(V_gen, "Found extention (%u) for the %s file\n",
	min_timet, spool_name[num]);

  if(f->fd)		/* close the previous file */
    close_spool_file(sph, num);

  /* Create full filename */
  if(snprintf(f->filepath, MAXPATHLEN, "%s/%s",
	      sph->spool_dir, fname) >= MAXPATHLEN) {
    logprint("Spool %s filename buffer overflow\n", spool_name[num]);
    return 0;
  }

  /* stat the file */
  if((stat(f->filepath, &statbuf)) != 0) {
    logerr("stat", f->filepath);
    return 0;
  }

  /* open */
  if((f->fd = open(f->filepath, O_RDONLY)) == -1) {
    logerr("open", f->filepath);
    return 0;
  }

  f->size = statbuf.st_size;
  f->inode = statbuf.st_ino;
  f->record = f->offset = 0;
  f->timet = min_timet;

  logprint("using %s file \"%s\"\n",
      spool_name[num], f->filepath);

  return read_header(sph, num);
}

void delete_and_arch(char *filepath, int delete_processed, char *arch_dir)
{
  if(!filepath) return;

  if(arch_dir) {
    char buffer[MAXPATHLEN * 2];

    if(snprintf(buffer, sizeof(buffer), "/bin/cp %s %s",
		filepath, arch_dir) >= sizeof(buffer))
      logprint("Archive buffer overflow\n");

    else if(system(buffer) == 0)
      logprint("file '%s' is archived to '%s'...\n",
	  filepath, arch_dir);
  }

  if(delete_processed) {
    if(unlink(filepath) == 0)
      logprint("file \"%s\" is deleted.\n", filepath);
    else logprint("unlink \"%s\": %s\n", filepath, strerror(errno));
  }
}

static void close_spool_file(spool_t *sph, int num)
{
  spoolfile_t *f;

  assert(sph);

  debug(V_func, "-> close_spool_file()\n");

  f = &sph->file[num];

  if(f->fd) {	    /* if no file was opened, touch nothing on disk.*/
    close(f->fd);
    f->fd = 0;

    logprint("closing %s file '%s': %d record(s) processed.\n",
	spool_name[num], f->filepath, f->record);

    delete_and_arch(f->filepath, sph->delete_processed, sph->arch_dir);
  }

  memset(f, 0, sizeof(spoolfile_t));
}

/* returns:
 -1: no more data available at this moment
  0: error
  1: success
*/
static int get_alert(UnifiedAlertRecord *uar, spool_t *sph)
{
  int ret;

  assert(uar);
  assert(sph);

  debug(V_all, "-> get_alert()\n");

  ret = read_piece((char *)uar, sizeof(UnifiedAlertRecord), sph, AFILE, 1);

  if(ret == 1)
    sph->alert_record ++;

  return ret;
}


/* returns:
 -1: no more data available at this moment
  0: error
  1: success
*/
static int get_log(UnifiedLogRecord *ulr, spool_t *sph, char *packet, int p_len)
{
  int ret;

  assert(ulr);
  assert(sph);

  debug(V_all, "-> get_log()\n");

  ulr->pkt = NULL; /* this value will be changed upon successful read operation */

  if((ret = read_piece((char *)&ulr->log, sizeof(UnifiedLog), sph, LFILE, 1)) == 1) {
    int packet_len = ulr->log.pkth.caplen;

    if(packet_len >= p_len) { /* large packet. Ignore it. */

      if((sph->log_offset = lseek(sph->log_fd, packet_len, SEEK_CUR)) == -1)
	child_fatal("lseek \"%s\": %s\n", sph->log_filepath, strerror(errno));

      logprint("'%s': skipping large packet (%d bytes)\n",
		 sph->log_filepath, packet_len);

      packet = NULL;
      ret = 1;
    }
    else 
      ret = read_piece(packet, packet_len, sph, LFILE, 0);
  }

  if(ret == 1) {
    sph->log_record++;
    ulr->pkt = packet;
  }

  return ret;
}

/* opens a spool file, if it is not opened yet by calling find_spool_file()
   reads 'len' bytes from the spool file to a 'buffer'
   if open_new > 0, calls find_spool_file() in the case of EOF
   returns:
   -1: no more data available at this moment
   0: error
   1: success
*/
static int read_piece(char *buffer, int len, spool_t *sph, int num, int open_new)
{
#define FIL   sph->file[num]

  size_t bytes_read = 0;
  int success = 0;
  int err_count = 0;
  int ret;

  assert(buffer);
  assert(sph);

  debug(V_all, "-> read_piece()\n");

  if(FIL.fd == 0)		/* trying to open file */
    if((ret = find_spool_file(sph, num)) != 1) return ret;

  while(!success && err_count++ < 3) { /* 3 tries to make it */

    if((bytes_read = read(FIL.fd, buffer, len)) == len)
      success = 1;

    else if(bytes_read == 0) {	/* EOF */
      if(open_new)
	switch (ret = find_spool_file(sph, num)) {
	case 1: err_count --;	/* new file is found: read again */
	  break;
	default: return ret;
	}
    }

    else {
      if(bytes_read == -1) /* error. unspecified position change might occur. */
	logerr("read", FIL.filepath);
      else			/* partial read */
	logprint("partial read (%d out of %d) on file %s",
		   bytes_read, len, FIL.filepath);

      if(lseek(FIL.fd, FIL.offset, SEEK_SET) == -1)
	child_fatal("lseek \"%s\": %s", FIL.filepath, strerror(errno));

      my_sleep();	/* relax and try again */
    }
  }

  if(success) {
    FIL.offset += bytes_read;
    return 1;
  }
  else return 0;

#undef FIL
}

static int save_check(spool_t *sph)
{
  static int size = sizeof(spoolfile_t) * 2;
  int fd;

  assert(sph);
  assert(sph->check_file);

  debug(V_all, "-> save_check()\n");

  /* return success if there is nothing to save */
  if(!*sph->alert_filepath && !*sph->log_filepath)
    return 1;

  if((fd = creat(sph->check_file, 0600)) == -1) {
    logerr("open", sph->check_file);
    return 0;
  }

  if(write(fd, &sph->file, size) != size) {
    logprint("write", sph->check_file);
    return 0;
  }

  close(fd);

  return 1;
}

static void restore_check(spool_t *sph)
{
  int fd;
  int i;
  spoolfile_t tmp[2];

  assert(sph);
  assert(sph->check_file);

  debug(V_func, "-> restore_check()\n");

  if((fd = open(sph->check_file, O_RDONLY)) == -1) {
    if(errno != ENOENT)
      logerr("open", sph->check_file);
    return;
  }

  if(read(fd, &tmp, sizeof(tmp)) != sizeof(tmp)) {
    logprint("read", sph->check_file);
    return;
  }

  close(fd);

  for(i = 0; i < 2; i++) {
    if(!*tmp[i].filepath) continue;

    if((tmp[i].fd = open(tmp[i].filepath, O_RDONLY)) == -1) {
      logerr("open", tmp[i].filepath);
      continue;
    }

    /* file is opened: check the magic number */
    if(!check_magic("", tmp[i].filepath, i)) continue;

    if(lseek(tmp[i].fd, tmp[i].offset, SEEK_SET) != tmp[i].offset) {
      logerr("lseek", tmp[i].filepath);
      continue;
    }

    /* we are good to go */
    memcpy(&sph->file[i], &tmp[i], sizeof(spoolfile_t));

    logprint("continue on %s file \"%s\": offset %u\n",
	spool_name[i], tmp[i].filepath, tmp[i].offset);

  } /* for() */
}
